

class MinimumTokenLimit(Exception):
    """
    The specified minimum token count has been exceeded
    """
    
class MaximumTokenLimit(Exception):
    """
    The specified maximum token count has been exceeded
    """



class ExtractException(Exception):
    """
    various extraction errors
    """
    
class UnsupportedHostException(ExtractException):
    """
    The URL is for a host we know we can't access reliably
    """

class UnsupportedContentType(ExtractException):
    """
    The http content type is unsupported.
    """
    
class EmptyText(ExtractException):
    """
    Unable to extract any readable text from the URL.
    """


class NetworkError(ExtractException):
    """
    Unable to access the content.
    """

    



"""
retry decorator in a wsk style
"""
from loguru import logger
from time import sleep
from functools import wraps

def retry(ExceptionToCheck=Exception, tries=5, delay=0.5, backoff=2, ExceptionToRaise=AssertionError):
    """Retry calling the decorated function using an exponential backoff.

    http://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/
    original from: http://wiki.python.org/moin/PythonDecoratorLibrary#Retry

    :param ExceptionToCheck: the exception to check. may be a tuple of
        exceptions to check
    :type ExceptionToCheck: Exception or tuple
    :param tries: number of times to try (not retry) before giving up
    :type tries: int
    :param delay: initial delay between retries in seconds
    :type delay: int
    :param backoff: backoff multiplier e.g. value of 2 will double the delay
        each retry
    :type backoff: int
    :param ExceptionToRaise: exceptions that should be raised instead of retried.
        
    """
    def deco_retry(f):

        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except ExceptionToRaise as e:
                    raise e
                except ExceptionToCheck as e:
                    if mtries == 1: # only show full stack trace on last try
                        logger.exception(f"Exception: {e}")
                    logger.warning(f"Exception: {e}")
                    logger.info(f"retrying in {mdelay} seconds")
                    sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return f(*args, **kwargs)

        return f_retry  # true decorator

    return deco_retry


from io import BytesIO 
from pdfminer.high_level import extract_pages
from pdfminer.layout import LTTextContainer


def pdf_text(pdf_bytes):
    """
    extract text from the pdf_bytes and return it as a single string
    """
    text = ""
    for page_layout in extract_pages(BytesIO(pdf_bytes)):
        for element in page_layout:
            if isinstance(element, LTTextContainer):
                for text_line in element:
                    text += text_line.get_text().rstrip() + " "
    return text



# wrangle text out of github repo readme via api
# only works for repo readme, not other github pages like issues, discussions, etc

from loguru import logger
import requests
import markdown 
from bs4 import BeautifulSoup 


def md_to_text(md):
    html = markdown.markdown(md)
    soup = BeautifulSoup(html, features='html.parser')
    return soup.get_text()


def github_readme_text(github_repo_url):
    # split a github url into the owner and repo components
    # use the github api to try to find the readme text
    # ['https:', '', 'github.com', 'jiggy-ai', 'hn_summary']
    spliturl = github_repo_url.rstrip('/').split('/')
    if len(spliturl) != 5:
        logger.warning(github_repo_url)
        raise Exception(f"Unable to process github url {github_repo_url}")
    owner = spliturl[3]
    repo = spliturl[4]
    contenturl = f'https://api.github.com/repos/{owner}/{repo}/readme'
    resp  = requests.get(contenturl)
    if resp.status_code != 200:
        logger.warning(f"{github_repo_url} {resp.content}")
        raise Exception(f"Unable to get readme for {github_repo_url}")
    item = resp.json()
    md = requests.get(item['download_url']).text
    return md_to_text(md), f'{owner}/{repo}'




"""
Extract readable text from a URL via various hacks
"""

from loguru import logger
from bs4 import BeautifulSoup, NavigableString, Tag
from readability import Document    # https://github.com/buriy/python-readability
import requests
import urllib.parse

from github_api import github_readme_text
from pdf_text import pdf_text

from exceptions import *
    
user_agent = "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/111.0.0.0 Safari/537.36"

headers = {'User-Agent': user_agent}

def extract_text_from_html(content):
    soup = BeautifulSoup(content, 'html.parser')

    blacklist = ['[document]','noscript','header','html','meta','head','input','script', "style"]
    # there may be more elements we don't want

    output =  ""      
    for t in soup.find_all(text=True):
        if t.parent.name not in blacklist:
            output += '{} '.format(t)
    return output

def get_language(html):
    soup = BeautifulSoup(html, 'html.parser')
    try:
        return soup.html["lang"]
    except:
        return ""

            

from retry import retry
    
@retry(tries=5)    
def get_url_text(url):
    """
    get url content and extract readable text
    returns the text
    """
    resp = requests.get(url, headers=headers, timeout=30)

    if resp.status_code != 200:
        logger.warning(url)
        raise NetworkError(f"Unable to get URL ({resp.status_code})")

    CONTENT_TYPE = resp.headers['Content-Type']
    
    if 'pdf' in CONTENT_TYPE:
        text, title, language = pdf_text(resp.content)
        return text, title, language    
    
    if "html" not in CONTENT_TYPE:
        logger.warning(url)
        raise UnsupportedContentType(f"Unsupported content type: {resp.headers['Content-Type']}")

    language = get_language(resp.text)
    logger.info(f"language: {language}")
        
    doc = Document(resp.text)
    title = doc.title()
    text = extract_text_from_html(doc.summary())

    if not len(text) or text.isspace():
        logger.warning(url)
        raise EmptyText("Unable to extract text data from url")
    return text, title, language



def url_to_text(url):
    logger.info("url_to_text: "+url)
    HOPELESS = ["youtube.com",
                "www.youtube.com"]
    if urllib.parse.urlparse(url).netloc in HOPELESS:
        logger.warning(url)        
        raise UnsupportedHostException("Unsupported host: {urllib.parse.urlparse(url).netloc}")

    if urllib.parse.urlparse(url).netloc == 'github.com':
        # for github repos use api to attempt to find a readme file
        text, title = github_readme_text(url)
        language = 'en'  # XXX  dynamically determine language
    else:
        text, title, language = get_url_text(url)

    #logger.debug("url_to_text: "+text)
    return text, title, language


if __name__ == "__main__":
    import sys
    url = sys.argv[-1]    
    from extract import url_to_text

    text, title, language = url_to_text(url)
    print(text)    
    print(f'language: {language}')
    print(f'Title: {title}')
